using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// HZMCUACOLLISIONS v1.0
//
// Improve CUA collisions when you need to interact with it

namespace VAMAAmbiencePlugin
{
    public class VAMAAmbience : MVRScript
    {	
        public static string PLUGIN_PATH;


		public Transform CUARoot;
		public Transform AmbienceRoot;
		private GameObject GoADebugMaxDist;
		public AudioSource AsAmbience;
		public AudioLowPassFilter AlpfAmbience;
		
		private bool cuaInitialized = false;
		
		// VAM UI	
		public JSONStorableStringChooser Ambience;
		protected List<string> ambienceChoices;
		
		public JSONStorableFloat AmbienceMaxDist;
		public JSONStorableFloat AmbienceVolume;
		public JSONStorableFloat AmbiencePitch;
		public JSONStorableFloat AmbienceSpatialization;
		public JSONStorableFloat AmbienceReverb;
		public JSONStorableFloat AmbienceLowPass;
		
		public JSONStorableBool ShowDebug;
					
		// Security
		private int initLoops = 0;
		private int maxInitLoops = 50;
		
		// Assetbundle
		private Request audioBundleRequest = null;
		
		// Others
		private List<string> ambiencesPaths;
		
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add VAMAAmbience on a Custom Unity Asset Atom");
                    return;
                }
				
				PLUGIN_PATH = GetPluginPath(this);			

				// Ambiences Choises
				ambienceChoices = new List<string>();
				
				Ambience = new JSONStorableStringChooser("Ambience", ambienceChoices, "", "Ambience", AmbienceCallback);
				UIDynamicPopup FAdp = CreateScrollablePopup(Ambience, false);
				FAdp.labelWidth = 180f;
				
				AmbienceMaxDist = new JSONStorableFloat("Ambience max distance", 25f, AmbienceMaxDistCallback, 1f, 100f);
				CreateSlider(AmbienceMaxDist, false);
				
				AmbienceVolume = new JSONStorableFloat("Ambience volume", 0.75f, AmbienceVolumeCallback, 0f, 1f);
				CreateSlider(AmbienceVolume, false);

				AmbiencePitch = new JSONStorableFloat("Ambience pitch", 1f, AmbiencePitchCallback, -3f, 3f);
				CreateSlider(AmbiencePitch, false);
				
				AmbienceSpatialization = new JSONStorableFloat("Ambience spatialization", 1f, AmbienceSpatializationCallback, 0f, 1f);
				CreateSlider(AmbienceSpatialization, false);
				
				AmbienceReverb = new JSONStorableFloat("Ambience reverb", 1f, AmbienceReverbCallback, 0f, 1.1f);
				CreateSlider(AmbienceReverb, false);
			
				AmbienceLowPass = new JSONStorableFloat("Ambience low pass", 22000f, (val) => { AmbienceLowPass.valNoCallback = Mathf.Round(val); AmbienceLowPassCallback( AmbienceLowPass.val ); } , 0f, 22000f);
				CreateSlider(AmbienceLowPass, false);
				
				ShowDebug = new JSONStorableBool("Show bounds", false, ShowDebugCallback);
				CreateToggle(ShowDebug, false);
			
				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help - Ambience</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>IMPORTANT:</b> If you switch asset in the Asset tab of the CUA, don't forget to reload the script to enable customization.\n\n" +
					"<b>Ambience max distance:</b> the maximum distance at which the ambience can be heard. You can check <b>Show bounds</b> below to see a gizmo previewing the distance.\n\n" +
					"<b>Ambience volume:</b> the volume of the ambience.\n\n" +
					"<b>Ambience pitch:</b> the pitch of the ambience. A negative value will play the ambience in reverse.\n\n" +
					"<b>Ambience spatialization:</b> if the sound is coming from its position in the world (3D / 5.1) or if it can be heard from anywhere (2D / Stereo).\n\n" +
					"<b>Ambience reverb:</b> how much the ambience sends its signal to reverb zones around it. You would have to use another CUA with the reverb zone or else this settings will have no apparent effect.\n\n" +
					"<b>Ambience low pass:</b> how much the ambience cuts the high frequencies. At 22000 you can hear all the frequencies. At 0 you will hear nothing. It is useful to simulate a <i>muffled effect</i>.\n\n" +
					"<b>Show Bounds:</b> is a visual gizmo to help you tweak the maximum distance of the ambience.\n" +
					"</color>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1150.0f;
			
				RegisterStringChooser(Ambience);
				RegisterFloat(AmbienceMaxDist);
				RegisterFloat(AmbienceVolume);
				RegisterFloat(AmbiencePitch);
				RegisterFloat(AmbienceSpatialization);
				RegisterFloat(AmbienceReverb);
				RegisterFloat(AmbienceLowPass);
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAAmbience - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }
		
		void Update() {
			// If we haven't made over 1min of init tests, we aren't initialized and the particle system (CUA) is still compatible, try to initialize
			if( initLoops < maxInitLoops && cuaInitialized == false ) {
				initLoops++;
				InitCUA();
			}
			
			if( initLoops == maxInitLoops ) {
				logError("VAMAAmbience : CUA could not be initialized properly after 3600 tries. Please select a compatible CUA and reload the script.");
				initLoops++; // To avoid spamming the log
			}
		}


		// **************************
		// Functions
		// **************************
				
		private bool InitCUA() {
			if( cuaInitialized == true ) return true;
			
			CUARoot = getCUARoot( containingAtom.reParentObject );
			if( CUARoot == null ) {
				return false;
			}
			
			AmbienceRoot = getObjectNamed( CUARoot, "ambience_audio");
			if( AmbienceRoot == null ) {
				return false;
			}
						
			AsAmbience = AmbienceRoot.gameObject.GetComponent<AudioSource>();
			if( AsAmbience != null ) {
				AlpfAmbience = AmbienceRoot.gameObject.GetComponent<AudioLowPassFilter>();
				
				Transform TrDebug = getObjectNamed( CUARoot, "debug_radius_max");
				GoADebugMaxDist = TrDebug.gameObject;
				
				cuaInitialized = true;
				logDebug("VAMAAmbience : Init successful");

				// Loading the bundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + "/vamatmosphere_ambiences.audiobundle", callback = OnAudioBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
			}
			return false;

		}
		
		protected void InitPaths() {
			// Creating assets paths list
			ambiencesPaths = new List<string>();
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_bar_south_africa.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_bar_us.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_birds_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_city_birds_dogs_cars.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_city_night.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_ethereal_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_forest_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_forest_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_night_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_night_strange_jungle.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_ocean_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_party_germany.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_river_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rumble_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_rumble_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scary_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scary_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scary_03.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scifi_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_scifi_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_city_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_seaside_city_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_underwater_01.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_underwater_02.ogg");
			ambiencesPaths.Add("Assets/VAMAtmosphere/audio_ambiences/amb_wind_01.ogg");
		}
		
		private void UpdateAmbienceSettings() {
			AmbienceCallback( Ambience.val );
			
			AmbienceMaxDistCallback(AmbienceMaxDist.val);
			AmbienceVolumeCallback(AmbienceVolume.val);
			AmbiencePitchCallback(AmbiencePitch.val);
			AmbienceSpatializationCallback(AmbienceSpatialization.val);
			AmbienceReverbCallback(AmbienceReverb.val);
			AmbienceLowPassCallback(AmbienceLowPass.val);
			
			ShowDebugCallback( ShowDebug.val );
		}
		
		protected void UpdateDebugDisplay() {
			if ( GoADebugMaxDist != null ) {
				GoADebugMaxDist.transform.localScale = new Vector3( AmbienceMaxDist.val, AmbienceMaxDist.val, AmbienceMaxDist.val );
			}
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		private void OnAudioBundleLoaded(Request aRequest) {
			try
            {
				InitPaths();
				
				audioBundleRequest = aRequest;
				
				ambienceChoices = new List<string>();
				foreach( string path in ambiencesPaths ) {
					string[] aPath = path.Split('/');
					string[] aFile = aPath.Last().Split('.');
					string fileName = aFile[aFile.Length - 2].ToString();
					ambienceChoices.Add( fileName );
				}
				
				Ambience.choices = ambienceChoices;
				
				UpdateAmbienceSettings();
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAAmbience - Exception caught: " + e);
            }
		}
		
		protected void AmbienceCallback(string choice) {
			if( CUARoot && AsAmbience ) {
				// I need to prevent a callback if the lists are not ready
				if( ambiencesPaths != null && ambienceChoices != null && ambiencesPaths.Count > 0 && ambienceChoices.Count > 0 && choice != "" ) {
					try{
						AudioClip ac = audioBundleRequest.assetBundle.LoadAsset<AudioClip>(ambiencesPaths[ ambienceChoices.IndexOf(choice) ]);
						AsAmbience.clip = ac;
						
						// Defining the start point of the sound to avoid chorus / echo effect (and also randomness!)
						float startTime = UnityEngine.Random.Range(0, Mathf.Round( ac.length / 2 ));
						AsAmbience.time = startTime;
						
						AsAmbience.Play();
					}
					catch(Exception e)
					{
						SuperController.LogError("VAMAAmbience - Exception caught: " + e);
					}
				}
			}
		}
		
		protected void AmbienceMaxDistCallback(float size) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.maxDistance = size;
				
				UpdateDebugDisplay();
			}
		}
		
		protected void AmbienceVolumeCallback(float volume) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.volume = volume;
			}
		}


		protected void AmbiencePitchCallback(float pitch) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.pitch = pitch;
			}
		}
		
		protected void AmbienceSpatializationCallback(float val) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.spatialBlend = val;
			}
		}
		
		protected void AmbienceReverbCallback(float val) {
			if( CUARoot && AsAmbience ) {
				AsAmbience.reverbZoneMix = val;
			}
		}
		
		protected void AmbienceLowPassCallback(float freq) {
			if( CUARoot && AlpfAmbience ) {
				AlpfAmbience.cutoffFrequency = freq;
			}
		}
		
		protected void ShowDebugCallback(bool enabled) {
			if( CUARoot ) {				
				if( enabled == true ) {
					GoADebugMaxDist.SetActive(true);			
					
					UpdateDebugDisplay();
				} else {
					GoADebugMaxDist.SetActive(false);
					
					UpdateDebugDisplay();
				}
			}
		}
			
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
        // Recursive function searching for a transform
        private Transform getCUARoot( Transform parent ) {
            foreach( Transform child in parent )
            {
				if( child.name.Contains("VAMAAmbience") ) {
					return child;
				} else {
					Transform childSearch = getCUARoot(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		private Transform getObjectNamed( Transform parent, string objName ) {
            foreach( Transform child in parent )
            {
				if( child.name == objName ) {
					return child;
				} else {
					Transform childSearch = getObjectNamed(child,objName);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnEnable() {
			// Disabling debug
			ShowDebugCallback( ShowDebug.val );
			
		}
		
		void OnDisable() {
			// Disabling debug
			ShowDebugCallback( false );
		}
		
		void OnDestroy() {
			// Disabling debug
			ShowDebugCallback( false );
			
			// Cleaning modified native VAM assets
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + "/vamatmosphere_ambiences.audiobundle");

		}

		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		

	}
}
