using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;

// HZMCUACOLLISIONS v1.0
//
// Improve CUA collisions when you need to interact with it

namespace VAMAReverbZonePlugin
{
    public class VAMAReverbZone : MVRScript
    {
		public Transform CUARoot;
		
		// Game Objects and components
		private GameObject GoRZDebugRadiusMin;
		private GameObject GoRZDebugRadiusMax;
		private AudioReverbZone RzComponent;
		
		private bool cuaInitialized = false;
		
		// VAM UI
		public JSONStorableFloat ZoneSizeMin;
		public JSONStorableFloat ZoneSizeMax;
		
		public JSONStorableStringChooser ZonePreset;
		protected List<string> presetsChoices;
		protected List<AudioReverbPreset> presetsList;
		
		public JSONStorableBool ShowDebug;
			
		// Security
		private int initLoops = 0;
		private int maxInitLoops = 3600;
		
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "CustomUnityAsset")
                {
                    SuperController.LogError("Please add VAMAReverbZone on a Custom Unity Asset Atom");
                    return;
                }
				
				ZoneSizeMin = new JSONStorableFloat("Zone size min", 10f, (val) => { ZoneSizeMinCallback( ZoneSizeMin.val ); ZoneSizeMax.SetVal( Mathf.Max(ZoneSizeMax.val, val) ); }, 1f, 100f);
				CreateSlider(ZoneSizeMin, false);
				
				ZoneSizeMax = new JSONStorableFloat("Zone size max", 15f, (val) => { ZoneSizeMaxCallback( ZoneSizeMax.val ); ZoneSizeMin.SetVal( Mathf.Min(ZoneSizeMin.val, val) ); }, 1f, 100f);
				CreateSlider(ZoneSizeMax, false);
				
				// Reverb Choises
				presetsChoices = new List<string>();
				presetsChoices.Add("Generic");
				presetsChoices.Add("PaddedCell");
				presetsChoices.Add("Room");
				presetsChoices.Add("Bathroom");
				presetsChoices.Add("Livingroom");
				presetsChoices.Add("Stoneroom");
				presetsChoices.Add("Auditorium");
				presetsChoices.Add("Concerthall");
				presetsChoices.Add("Cave");
				presetsChoices.Add("Arena");
				presetsChoices.Add("Hangar");
				presetsChoices.Add("CarpetedHallway");
				presetsChoices.Add("Hallway");
				presetsChoices.Add("StoneCorridor");
				presetsChoices.Add("Alley");
				presetsChoices.Add("Forest");
				presetsChoices.Add("City");
				presetsChoices.Add("Mountains");
				presetsChoices.Add("Quarry");
				presetsChoices.Add("Plain");
				presetsChoices.Add("ParkingLot");
				presetsChoices.Add("SewerPipe");
				presetsChoices.Add("Underwater");
				presetsChoices.Add("Drugged");
				presetsChoices.Add("Dizzy");
				presetsChoices.Add("Psychotic");
				
				presetsList = new List<AudioReverbPreset>();
				presetsList.Add(AudioReverbPreset.Generic);
				presetsList.Add(AudioReverbPreset.PaddedCell);
				presetsList.Add(AudioReverbPreset.Room);
				presetsList.Add(AudioReverbPreset.Bathroom);
				presetsList.Add(AudioReverbPreset.Livingroom);
				presetsList.Add(AudioReverbPreset.Stoneroom);
				presetsList.Add(AudioReverbPreset.Auditorium);
				presetsList.Add(AudioReverbPreset.Concerthall);
				presetsList.Add(AudioReverbPreset.Cave);
				presetsList.Add(AudioReverbPreset.Arena);
				presetsList.Add(AudioReverbPreset.Hangar);
				presetsList.Add(AudioReverbPreset.CarpetedHallway);
				presetsList.Add(AudioReverbPreset.Hallway);
				presetsList.Add(AudioReverbPreset.StoneCorridor);
				presetsList.Add(AudioReverbPreset.Alley);
				presetsList.Add(AudioReverbPreset.Forest);
				presetsList.Add(AudioReverbPreset.City);
				presetsList.Add(AudioReverbPreset.Mountains);
				presetsList.Add(AudioReverbPreset.Quarry);
				presetsList.Add(AudioReverbPreset.Plain);
				presetsList.Add(AudioReverbPreset.ParkingLot);
				presetsList.Add(AudioReverbPreset.SewerPipe);
				presetsList.Add(AudioReverbPreset.Underwater);
				presetsList.Add(AudioReverbPreset.Drugged);
				presetsList.Add(AudioReverbPreset.Dizzy);
				presetsList.Add(AudioReverbPreset.Psychotic);
				
				ZonePreset = new JSONStorableStringChooser("Reverb preset", presetsChoices, "Generic", "Reverb preset", ZonePresetCallback);
				UIDynamicPopup FAdp = CreateScrollablePopup(ZonePreset, false);
				FAdp.labelWidth = 230f;
				
				ShowDebug = new JSONStorableBool("Show bounds", false, ShowDebugCallback);
				CreateToggle(ShowDebug, false);

				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Help - Reverb zone</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>IMPORTANT:</b> If you switch asset in the Asset tab of the CUA, don't forget to reload the script to enable customization.\n\n" +
					"<b>Zone size min:</b> is the area where you can hear the reverb effect at full volume.\n\n" +
					"<b>Zone size max:</b> is the area where the reverb fades out. Having a bit of distance from the min makes the transition smoother when you move away from the reverb zone.\n\n" +
					"<b>Reverb preset:</b> is the way the reverb sounds.\n\n" +
					"<b>Show bounds:</b> is a visual gizmo to help you tweak the reverb zone.\n" +
					"</color>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, true);
				helpTextfield.height = 1150.0f;
				
				RegisterFloat(ZoneSizeMin);
				RegisterFloat(ZoneSizeMax);
				RegisterStringChooser(ZonePreset);			
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAReverbZone - Exception caught: " + e);
            }
		}
		
		void Start()
        {

        }
		
		void Update() {
			// If we haven't made over 1min of init tests, we aren't initialized and the particle system (CUA) is still compatible, try to initialize
			if( initLoops < maxInitLoops && cuaInitialized == false ) {
				initLoops++;
				InitCUA();
			}
			
			if( initLoops == maxInitLoops ) {
				logError("VAMAReverbZone : CUA could not be initialized properly after 3600 tries. Please select a compatible CUA and reload the script.");
				initLoops++; // To avoid spamming the log
			}
		}
		

		
		// **************************
		// Functions
		// **************************
				
		private bool InitCUA() {
			if( cuaInitialized == true ) return true;
			
			CUARoot = getCUARoot( containingAtom.reParentObject );
			if( CUARoot == null ) {
				return false;
			}
			
			RzComponent = CUARoot.gameObject.GetComponent<AudioReverbZone>();
			if( RzComponent != null ) {
				
				foreach( Transform child in CUARoot ) {
					if( child.name == "debug_radius_min" ) GoRZDebugRadiusMin = child.gameObject;
					if( child.name == "debug_radius_max" ) GoRZDebugRadiusMax = child.gameObject;
				}
				
				cuaInitialized = true;
				logDebug("VAMAReverbZone : Init successful");
				
				UpdateCUASettings();
			}
		
			return false;

		}
		
		protected void UpdateCUASettings() {
			ZoneSizeMinCallback(ZoneSizeMin.val);
			ZoneSizeMaxCallback(ZoneSizeMax.val);
			ZonePresetCallback(ZonePreset.val);
		}
		
		protected void UpdateDebugDisplay() {
			if ( GoRZDebugRadiusMin != null && GoRZDebugRadiusMax != null ) {
				GoRZDebugRadiusMin.transform.localScale = new Vector3( ZoneSizeMin.val, ZoneSizeMin.val, ZoneSizeMin.val );
				GoRZDebugRadiusMax.transform.localScale = new Vector3( ZoneSizeMax.val, ZoneSizeMax.val, ZoneSizeMax.val );
			}
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		protected void ZoneSizeMinCallback(float size ) {
			if( CUARoot ) {
				RzComponent.minDistance = size;
				
				UpdateDebugDisplay();
			}
		}
		
		protected void ZoneSizeMaxCallback(float size ) {
			if( CUARoot ) {
				RzComponent.maxDistance = size;
				
				UpdateDebugDisplay();
			}
		}
		
		protected void ZonePresetCallback(string choice) {
			if( CUARoot ) {
				RzComponent.reverbPreset = presetsList[ presetsChoices.IndexOf(choice) ];
			}
		}
		
		
		protected void ShowDebugCallback(bool enabled) {
			if( CUARoot ) {				
				if( enabled == true ) {
					GoRZDebugRadiusMin.SetActive(true);
					GoRZDebugRadiusMax.SetActive(true);			
					
					UpdateDebugDisplay();
				} else {
					GoRZDebugRadiusMin.SetActive(false);
					GoRZDebugRadiusMax.SetActive(false);
					
					UpdateDebugDisplay();
				}
			}
		}
			
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
        // Recursive function searching for a transform
        private Transform getCUARoot( Transform parent ) {
            foreach( Transform child in parent )
            {
				if( child.name.Contains("VAMAReverbZone") ) {
					return child;
				} else {
					Transform childSearch = getCUARoot(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Cleaning modified native VAM assets

		}

	}
}